/**************************************************
*   Author: Morrison
*   Date:  12 Apr 202021
**************************************************/

import javafx.application.Application;
import javafx.application.Platform;
import javafx.stage.Stage;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.TextField;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.GridPane;
import java.util.function.DoubleBinaryOperator;
import java.util.function.DoubleUnaryOperator;
import javafx.geometry.Insets;

public class RPN extends Application
{
    private final TextField display;
    private final TextField stackShow;
    private NumberButton[] digits;
    private NumberButton decimalPoint;
    private NumberButton exponent;
    private Button clearButton;
    Button clearEntryButton;
    Button enterButton;
    Button backspaceButton;
    IStack<Double> theStack;
    //0p buttons
    private OpButton plus;
    private OpButton minus;
    private OpButton times;
    private OpButton divide;
    private OpButton power;
    //function buttons
    private FunctionButton sin;
    private FunctionButton cos;
    private FunctionButton tan;
    private FunctionButton asin;
    private FunctionButton acos;
    private FunctionButton atan;
    private Button pi;
    private Button degRadButton;
    private Mode mode;
    private boolean overwrite;
    public RPN()
    {
        display = new TextField();
        display.getStyleClass().add("textfield");
        stackShow = new TextField();
        stackShow.getStyleClass().add("textfield");
        digits = new NumberButton[10];
        for(int k = 0; k < 10; k++)
        {
            digits[k] = new NumberButton((char)(0x30 + k));
        }
        decimalPoint = new NumberButton('.');
        enterButton = new Button("ENTER");
        exponent = new NumberButton('E');
        theStack = new AStack<>();
        plus = new OpButton( (x,y) -> x + y   , "+");
        minus = new OpButton( (x,y) -> x - y, "-");
        times = new OpButton( (x,y) -> x*y, "*");
        divide = new OpButton( (x,y) -> x/y, "/");
        power = new OpButton( (x,y) -> Math.pow(x,y), "^");
        sin = new FunctionButton( x -> Math.sin(x*convert()), "sin");
        cos = new FunctionButton( x -> Math.cos(x*convert()), "cos");
        tan = new FunctionButton( x -> Math.tan(x*convert()), "tan");
        asin = new FunctionButton( x -> Math.asin(x)/convert(), "asin");
        acos = new FunctionButton( x -> Math.acos(x)/convert(), "acos");
        atan = new FunctionButton( x -> Math.atan(x)/convert(), "atan");
        pi = new Button("pi");
        clearButton = new Button("C");
        clearEntryButton = new Button("CE");
        backspaceButton = new Button("<=");
        degRadButton = new Button("RAD");
        mode = Mode.RADIANS;


        overwrite = false;
    }
    
    @Override
    public void init()
    {
    }

    @Override
    public void start(Stage primary)
    {
        BorderPane bp = new BorderPane();
        GridPane gp = new GridPane();
        gp.setHgap(10);
        gp.setVgap(10);
        gp.setPadding(new Insets(10));
        gp.add(digits[7], 0, 0);
        gp.add(digits[8], 1, 0);
        gp.add(digits[9], 2, 0);
        gp.add(digits[4], 0, 1);
        gp.add(digits[5], 1, 1);
        gp.add(digits[6], 2, 1);
        gp.add(digits[1], 0, 2);
        gp.add(digits[2], 1, 2);
        gp.add(digits[3], 2, 2);
        gp.add(decimalPoint, 0, 3);
        gp.add(digits[0], 1, 3);
        gp.add(exponent, 2, 3);
        gp.add(plus, 3, 0);
        gp.add(minus, 3, 1);
        gp.add(times, 3, 2);
        gp.add(divide, 3, 3);
        gp.add(power, 3, 4);
        gp.add(enterButton, 0, 4);
        gp.add(sin, 4, 0);
        gp.add(cos, 4, 1);
        gp.add(tan, 4, 2);
        gp.add(pi, 4,3);
        gp.add(backspaceButton, 5, 0);
        gp.add(clearEntryButton, 5, 1);
        gp.add(clearButton, 5, 2);
        gp.add(degRadButton, 5, 3);
        gp.add(asin, 6, 0);
        gp.add(acos, 6, 1);
        gp.add(atan, 6, 2);
        enterButton.setOnAction( e ->
        {
            //contenst of display go to the stack.
            //TODO: Handle end-user stupidity
            double contents = Double.parseDouble(display.getText());
            //display.setText("");
            overwrite = true;
            //push on stack
            theStack.push(contents);
            stackShow.setText(theStack.toString());
        });
        backspaceButton.setOnAction( e ->
        {
            String s = display.getText();
            if(s.length() > 0)
            {
                display.setText(s.substring(0, s.length() - 1));
            }
        });
        clearButton.setOnAction( e ->
        {
            display.setText("");
            theStack.clear();
            stackShow.setText(theStack.toString());
        });
        clearEntryButton.setOnAction( e ->
        {
            display.setText("");
        });
        degRadButton.setOnAction( e ->
        {
            if(mode == Mode.RADIANS)
            {
                degRadButton.setText("DEG");
                mode = Mode.DEGREES;
            }
            else
            {
                degRadButton.setText("RAD");
                mode = Mode.RADIANS;
            }
        });
        pi.setOnAction( e ->
        {
            display.setText("" + Math.PI);
        });
        bp.setTop(display);
        bp.setBottom(stackShow);
        bp.setCenter(gp);
        Scene scene = new Scene(bp, 700, 500);
        scene.getStylesheets().add("./RPN.css");
        primary.setTitle("NCSSM RPN Calculator Axe");
        primary.setScene(scene);
        primary.show();
    }

    @Override
    public void stop()
    {
    }
    private double convert()
    {
        return (mode == Mode.RADIANS)? 1: Math.PI/180;
    }  
    /********************** Inner Classes ********************/
    class NumberButton extends Button
    {
        private final char num;
        public NumberButton(char num)
        {
            super("" + num);
            this.num = num;
            setOnAction( e ->
            {
                if(overwrite)
                {
                    display.setText("" + num);
                }
                else
                {
                    display.setText(display.getText() + num);
                }
                overwrite = false;
            });
            getStyleClass().add("number-button");
        }

    }
    class OpButton extends Button
    {
        private DoubleBinaryOperator op;

        public OpButton(DoubleBinaryOperator op, String symbol)
        {
            super(symbol);
            this.op = op;
            setOnAction( e ->
            {
                double second = Double.parseDouble(display.getText());
                double first = theStack.pop();
                double result = op.applyAsDouble(first, second);
                //theStack.push(result);
                display.setText("" + result);
                stackShow.setText(theStack.toString());
                overwrite = true;
            });
        }
    }
    class FunctionButton extends Button
    {
        private DoubleUnaryOperator f;
        public FunctionButton(DoubleUnaryOperator f, String symbol)
        {
            super(symbol);
            this.f = f;
            setOnAction( e ->
            {
                double argument = Double.parseDouble(display.getText());
                double result = f.applyAsDouble(argument);
                display.setText(""  + result);
                overwrite = true;
            });
        }
    }
    enum Mode{RADIANS, DEGREES};
}
